import time
from decimal import Decimal
from struct import pack

from eth_hash.auto import keccak

from zex.sdk.client.signing_visitor import SigningVisitor
from zex.sdk.data_types import (
    CancelOrderRequest,
    OrderSide,
    PlaceOrderRequest,
    WithdrawRequest,
)


class SigningVisitorDev(SigningVisitor):
    def create_register_transaction(self) -> bytes:
        transaction_data = (
            pack(">B", self._version)
            + pack(">B", self._register_command)
            + pack(">B", self._signature_type.value)
            + self.public_key
        )
        signature = self._private_key.sign_recoverable(
            keccak(self._create_register_message()), hasher=None
        )
        signature = signature[:64]  # Compact format.
        transaction_data += signature
        return transaction_data

    def create_place_order_transaction(
        self, request: PlaceOrderRequest, nonce: int, user_id: int
    ) -> bytes:
        pair = request.base_token + request.quote_token

        price = round(Decimal(request.price), self._price_digits)
        volume = round(Decimal(request.volume), self._volume_digits)
        volume_mantissa, volume_exponent = self._to_scientific(volume)
        price_mantissa, price_exponent = self._to_scientific(price)

        volume = volume_mantissa * 10 ** Decimal(volume_exponent)
        price = price_mantissa * 10 ** Decimal(price_exponent)

        transaction_data = (
            pack(">B", self._version)
            + pack(
                ">B",
                (
                    self._buy_command
                    if request.side == OrderSide.BUY
                    else self._sell_command
                ),
            )
            + pack(">B", self._signature_type.value)
            + pack(">B", len(request.base_token))
            + pack(">B", len(request.quote_token))
            + pair.encode()
            + pack(">Q b", volume_mantissa, volume_exponent)
            + pack(">Q b", price_mantissa, price_exponent)
        )
        epoch = int(time.time())

        transaction_data += pack(">II", epoch, nonce) + pack(">Q", user_id)

        message = (
            "v: 1\n"
            f"name: {request.side.lower()}\n"
            f"base token: {request.base_token}\n"
            f"quote token: {request.quote_token}\n"
            f"amount: {self._format_decimal(volume)}\n"
            f"price: {self._format_decimal(price)}\n"
            f"t: {epoch}\n"
            f"nonce: {nonce}\n"
            f"user_id: {user_id}\n"
        )
        message = "\x19Ethereum Signed Message:\n" + str(len(message)) + message

        signature = self._private_key.sign_recoverable(
            keccak(message.encode("ascii")), hasher=None
        )
        signature = signature[:64]  # Compact format

        transaction_data += signature
        return transaction_data

    def create_cancel_order_transaction(
        self, request: CancelOrderRequest, user_id: int
    ) -> bytes:
        transaction_data = (
            pack(">B", self._version)
            + pack(">B", self._cancel_command)
            + pack(">B", self._signature_type.value)
            + pack(">Q", user_id)
            + pack(">Q", request.order_nonce)
        )

        message = (
            f"v: {transaction_data[0]}\n"
            "name: cancel\n"
            f"user_id: {user_id}\n"
            f"order_nonce: {request.order_nonce}\n"
        )
        message = "".join(
            ("\x19Ethereum Signed Message:\n", str(len(message)), message)
        )

        signature = self._private_key.sign_recoverable(
            keccak(message.encode("ascii")), hasher=None
        )
        signature = signature[:64]  # Compact format

        transaction_data += signature
        return transaction_data

    def create_withdraw_transaction(
        self, request: WithdrawRequest, nonce: int, user_id: int
    ) -> bytes:
        transaction_data = (
            pack(">B", self._version)
            + pack(">B", self._signature_type.value)
            + pack(">B", self._withdraw_command)
            + pack(">B", len(request.token_name))
            + request.token_chain.encode()
            + request.token_name.encode()
            + pack(">d", request.amount)
            + bytes.fromhex(request.destination[2:])
        )
        epoch = int(time.time())
        transaction_data += (
            pack(">II", epoch, nonce) + pack(">Q", user_id) + self.public_key
        )

        message = (
            "v: 1\n"
            "name: withdraw\n"
            f"token chain: {request.token_chain}\n"
            f"token name: {request.token_name}\n"
            f"amount: {request.amount}\n"
            f"to: {request.destination}\n"
            f"t: {epoch}\n"
            f"nonce: {nonce}\n"
            f"user_id: {user_id}\n"
        )
        message = "\x19Ethereum Signed Message:\n" + str(len(message)) + message

        signature = self._private_key.sign_recoverable(
            keccak(message.encode("ascii")), hasher=None
        )
        signature = signature[:64]  # Compact format

        transaction_data += signature
        return transaction_data

    @staticmethod
    def _to_scientific(number: Decimal) -> tuple[int, int]:
        """Convert a Decimal value to a mantissa and an exponent (base 10)."""

        sign, digits, exponent = number.normalize().as_tuple()
        if not isinstance(exponent, int):
            raise TypeError(f"Cannot convert value to scientific form: {number}")

        mantissa = 0
        for index, digit in enumerate(reversed(digits)):
            mantissa += digit * 10**index
        mantissa *= -1 if sign != 0 else 1

        if exponent < -128 or exponent > 127:
            raise RuntimeError(f"Cannot convert value to scientific form: {number}")

        return mantissa, exponent

    @staticmethod
    def _format_decimal(decimal_number: Decimal) -> str:
        """
        Format the given decimal number, making sure scientific notation \
        is *not* used and that there's always a traling .0 if there's no \
        fractional part.
        """
        result = format(decimal_number, "f")
        if "." not in result:
            result += ".0"
        return result



from pydantic import BaseModel, Field


class CancelOrderRequest(BaseModel):
    signed_order: bytes = Field(
        ..., description="The signature of the placed order to be cancelled."
    )
    order_nonce: int = Field(..., description="The nonce of the order to be cancelled.")



from .async_client import AsyncClient as AsyncClient
from .client import Client as Client
from .signing_visitor import SigningVisitor as SigningVisitor
from .signing_visitor_dev import SigningVisitorDev as SigningVisitorDev
from .signing_visitor_main import SigningVisitorMain as SigningVisitorMain



from __future__ import annotations

import asyncio
from collections.abc import Iterable
from enum import Enum
from typing import Any, TypeVar

import httpx
from pydantic import TypeAdapter

from zex.sdk.client.signing_visitor import SigningVisitor
from zex.sdk.client.signing_visitor_dev import SigningVisitorDev
from zex.sdk.client.signing_visitor_main import SigningVisitorMain
from zex.sdk.data_types import (
    Asset,
    CancelOrderRequest,
    Order,
    PlaceOrderRequest,
    PlaceOrderResult,
    TradeInfo,
    Transfer,
    Withdraw,
    WithdrawRequest,
)

ServerResponseType = TypeVar("ServerResponseType")


class SignatureType(Enum):
    SECP256K1 = 1
    ED25519 = 2


class AsyncClient:
    """
    The asynchronous client of Zex exchange supporting the main functionalities \
    of the exchange API.
    """

    def __init__(
        self,
        signing_visitor: SigningVisitor,
        testnet: bool = True,
    ) -> None:
        self._signing_visitor = signing_visitor
        self._api_endpoint = (
            "https://api-dev.zex.finance" if testnet else "https://api.zex.finance"
        )
        self._version = 1
        self.testnet = testnet

        self._register_timeout = 20.0
        self.nonce: int | None = None
        self.user_id: int | None = None

    @classmethod
    async def create(
        cls, api_key: str | None = None, testnet: bool = True
    ) -> AsyncClient:
        """
        Create a registered instance of asynchronous Zex client to use functionalities \
        provided by the Zex exchange.

        .. warning::
            The user of this class should only use this method to instantiate. The
            `__init__` method of this class is for internal use only.

        :param api_key: The API key used to register with Zex exchange servers and sign \
        the transactions.
        :param testnet: Whether or not to submit transactions into Zex testnet.
        """
        signing_visitor: SigningVisitor
        if testnet:
            signing_visitor = SigningVisitorDev(api_key=api_key)
        else:
            signing_visitor = SigningVisitorMain(api_key=api_key)
        client = cls(signing_visitor, testnet)
        await client.register_user_id()
        return client

    async def register_user_id(self) -> None:
        """
        Register the user ID with Zex exchange. Some methods the client to be registered \
        to be able to take effect in Zex server.

        .. note:
            The :py:meth:`~AsyncClient.create` method automatically calls this method and registers
            the client.
        """
        if self.user_id is not None:
            return

        transaction_data = self._signing_visitor.create_register_transaction()

        user_id = None
        async with httpx.AsyncClient() as client:
            await client.post(
                f"{self._api_endpoint}/v1/register",
                json=[transaction_data.decode("latin-1")],
                timeout=self._register_timeout,
            )
            try:
                user_id = await asyncio.wait_for(
                    self._fetch_user_id_from_server(client),
                    timeout=self._register_timeout,
                )
            except asyncio.TimeoutError as exc:
                raise RuntimeError("Registering user ID timed out.") from exc
        self.user_id = user_id

    async def place_batch_order(
        self, orders: Iterable[PlaceOrderRequest]
    ) -> Iterable[PlaceOrderResult]:
        """
        Place a batch of orders in Zex exchange.

        .. note:
            The client should be registered before calling this method.

        :param orders: The batch of orders to be placed in the exchange.
        """
        if self.user_id is None:
            raise RuntimeError("The Zex client is not registered.")

        orders = list(orders)

        if len(orders) == 0:
            return []

        async with httpx.AsyncClient() as client:
            nonce_response = await client.get(
                f"{self._api_endpoint}/v1/user/nonce?id={self.user_id}"
            )
            self.nonce = nonce_response.json()["nonce"]
            assert self.nonce is not None, "For typing."

        payload = []
        place_order_results = []
        for order in orders:
            signed_order_transaction = (
                self._signing_visitor.create_place_order_transaction(
                    order, self.nonce, self.user_id
                )
            )
            place_order_results.append(
                PlaceOrderResult(
                    place_order_request=order,
                    nonce=self.nonce,
                    signed_order_transaction=signed_order_transaction,
                )
            )
            self.nonce += 1
            payload.append(signed_order_transaction.decode("latin-1"))

        if not payload:
            return []

        async with httpx.AsyncClient() as client:
            await client.post(
                f"{self._api_endpoint}/v1/order",
                json=payload,
            )

        return place_order_results

    async def cancel_batch_order(
        self, cancel_orders: Iterable[CancelOrderRequest]
    ) -> None:
        """
        Cancel a batch of orders in Zex exchange.

        .. note:
            The client should be registered before calling this method.

        :param cancel_orders: The batch cancel order requests.
        """
        if self.user_id is None:
            raise RuntimeError("The Zex client is not registered.")

        payload = []
        for order in cancel_orders:
            cancel_order_transaction = (
                self._signing_visitor.create_cancel_order_transaction(
                    order, self.user_id
                )
            )
            payload.append(cancel_order_transaction.decode("latin-1"))
        if not payload:
            return
        async with httpx.AsyncClient() as client:
            await client.post(
                f"{self._api_endpoint}/v1/order",
                json=payload,
            )

    async def withdraw(self, withdraw_request: WithdrawRequest) -> None:
        """
        Withdraw from Zex exchange.

        .. note:
            The client should be registered before calling this method.

        :param withdraw_request: The withdraw request containing the required fields.
        """
        if self.user_id is None:
            raise RuntimeError("The Zex client is not registered.")

        async with httpx.AsyncClient() as client:
            nonce_response = await client.get(
                f"{self._api_endpoint}/v1/user/nonce?id={self.user_id}"
            )
            self.nonce = nonce_response.json()["nonce"]
            assert self.nonce is not None, "For typing."

        signed_withdraw_transaction = self._signing_visitor.create_withdraw_transaction(
            withdraw_request, self.nonce, self.user_id
        )
        payload = [signed_withdraw_transaction.decode("latin-1")]
        async with httpx.AsyncClient() as client:
            await client.post(
                f"{self._api_endpoint}/v1/withdraw",
                json=payload,
            )

    async def deposit(self, transaction: bytes) -> None:
        """
        Deposit in Zex exchange.

        .. note:
            The client should be registered before calling this method.

        :param transaction: The signed transaction for depositing in Zex exchange.
        """
        payload = [transaction.decode("latin-1")]
        async with httpx.AsyncClient() as client:
            await client.post(
                f"{self._api_endpoint}/v1/deposit",
                json=payload,
            )

    async def get_server_time(self) -> int:
        """Get the server time."""
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self._api_endpoint}/v1/time",
            )
        response_data: dict[str, int] = response.json()
        time = response_data.get("serverTime")
        if time is None:
            raise RuntimeError("The server did not return a proper response.")
        return time

    async def ping(self) -> bool:
        """Whether the server responds to a ping request."""
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self._api_endpoint}/v1/ping",
            )
        if response.status_code == 200:
            return True
        return False

    async def get_price(self, symbol: str) -> float:
        """
        Get the price of a market.

        :param symbol: The symbol of the market whose price to query.
        """
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self._api_endpoint}/v1/ticker/price",
                params={"symbol": symbol},
            )
        response_data = response.json()
        if response.status_code == 422:
            detail = response_data.get("detail") or []
            raise RuntimeError(f"Fetching price from the server failed: {detail}")
        price: float | None = response_data.get("price")
        if price is None:
            raise RuntimeError(
                "Could not retrieve the price of symbol from the server."
            )
        return price

    async def get_ticker(self, symbol: str) -> dict[str, Any]:
        """
        Get the ticker of a certain market.

        :param symbol: The symbol of the market to get the ticker of.
        """
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self._api_endpoint}/v1/ticker",
                params={"symbol": symbol},
            )
        response_data: dict[str, Any] = response.json()
        if response.status_code == 422:
            detail = response_data.get("detail") or []
            raise RuntimeError(f"Fetching ticker from the server failed: {detail}")
        return response_data

    async def get_depth(self, symbol: str, limit: int) -> dict[str, Any]:
        """
        Get the market depth.

        :param symbol: The symbol of the market to get the depth of.
        :param limit: The limit of the queried depth.
        """
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self._api_endpoint}/v1/depth",
                params={"symbol": symbol, "limit": limit},
            )
        response_data: dict[str, Any] = response.json()
        if response.status_code == 422:
            detail = response_data.get("detail") or []
            raise RuntimeError(
                f"Fetching market depth from the server failed: {detail}"
            )
        return response_data

    async def get_exchange_info(self, symbol: str) -> dict[str, Any]:
        """
        Get the exchange info for a specified market.

        :parma symbol: The symbol of the market to get the exchange info of.
        """
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self._api_endpoint}/v1/exchangeInfo",
                params={"symbol": symbol},
            )
        response_data: dict[str, Any] = response.json()
        if response.status_code == 422:
            detail = response_data.get("detail") or []
            raise RuntimeError(
                f"Fetching exchange info from the server failed: {detail}"
            )
        return response_data

    async def get_user_trades(self) -> list[TradeInfo]:
        """
        Get the list of all trades of the user.

        .. note:
            The client should be registered before calling this method.
        """
        if self.user_id is None:
            raise RuntimeError("The Zex client is not registered.")
        return await self._get_and_parse_response_from_server(
            type_adapter=TypeAdapter(list[TradeInfo]),
            api_path="/v1/user/trades",
            params={"id": self.user_id},
        )

    async def get_user_assets(self) -> list[Asset]:
        """
        Get the list of all assets of the user.

        .. note:
            The client should be registered before calling this method.
        """
        if self.user_id is None:
            raise RuntimeError("The Zex client is not registered.")
        return await self._get_and_parse_response_from_server(
            type_adapter=TypeAdapter(list[Asset]),
            api_path="/v1/asset/getUserAsset",
            params={"id": self.user_id},
        )

    async def get_user_orders(self) -> list[Order]:
        """
        Get the list of all the placed orders of the user.

        .. note:
            The client should be registered before calling this method.
        """
        if self.user_id is None:
            raise RuntimeError("The Zex client is not registered.")
        return await self._get_and_parse_response_from_server(
            type_adapter=TypeAdapter(list[Order]),
            api_path="/v1/user/orders",
            params={"id": self.user_id},
        )

    async def get_user_transfers(self) -> list[Transfer]:
        """
        Get the list of all transfers of the user.

        .. note:
            The client should be registered before calling this method.
        """
        if self.user_id is None:
            raise RuntimeError("The Zex client is not registered.")
        return await self._get_and_parse_response_from_server(
            type_adapter=TypeAdapter(list[Transfer]),
            api_path="/v1/user/transfers",
            params={"id": self.user_id},
        )

    async def get_user_withdraws(self, chain: str) -> list[Withdraw]:
        """
        Get the list of all withdraws of the user.

        .. note:
            The client should be registered before calling this method.
        """
        if self.user_id is None:
            raise RuntimeError("The Zex client is not registered.")
        return await self._get_and_parse_response_from_server(
            type_adapter=TypeAdapter(list[Withdraw]),
            api_path="/v1/user/withdraws",
            params={"id": self.user_id, "chain": chain},
        )

    async def _fetch_user_id_from_server(self, client: httpx.AsyncClient) -> int:
        while True:
            response = await client.get(
                f"{self._api_endpoint}/v1/user/id?public={self._signing_visitor.public_key.hex()}",  # noqa: E501
                timeout=self._register_timeout,
            )
            if response.status_code == 200:
                response_data = response.json()
                if "id" in response_data:
                    return int(response_data["id"])
            await asyncio.sleep(0.1)

    async def _get_and_parse_response_from_server(
        self,
        type_adapter: TypeAdapter[ServerResponseType],
        api_path: str,
        params: dict[str, Any] | None = None,
    ) -> ServerResponseType:
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self._api_endpoint}{api_path}",
                params=params,
            )

        response_data = response.json()
        if response.status_code == 422:
            detail = response_data.get("detail") or []
            raise RuntimeError(
                f"Fetching response from the server failed. Detail: {detail}"
            )

        try:
            return type_adapter.validate_python(response_data)
        except Exception as e:
            raise RuntimeError(f"Parsing response failed with error: {e}") from e



from abc import ABC, abstractmethod
from enum import Enum

from coincurve import PrivateKey

from zex.sdk.data_types import CancelOrderRequest, PlaceOrderRequest, WithdrawRequest


class SignatureType(Enum):
    SECP256K1 = 1
    ED25519 = 2


class SigningVisitor(ABC):
    def __init__(
        self,
        api_key: str | None = None,
    ) -> None:
        private_key_bytes = bytes.fromhex(api_key) if api_key is not None else None
        self._private_key = PrivateKey(secret=private_key_bytes)
        self.public_key = self._private_key.public_key.format(compressed=True)

        self._version = 1
        self._signature_type = SignatureType.SECP256K1

        self._register_command = ord("r")
        self._buy_command = ord("b")
        self._sell_command = ord("s")
        self._cancel_command = ord("c")
        self._withdraw_command = ord("w")
        self._deposit_command = ord("d")
        self._btc_deposit_command = ord("x")

        self._price_digits = 2
        self._volume_digits = 5

    @abstractmethod
    def create_register_transaction(self) -> bytes:
        pass

    @abstractmethod
    def create_place_order_transaction(
        self, request: PlaceOrderRequest, nonce: int, user_id: int
    ) -> bytes:
        pass

    @abstractmethod
    def create_cancel_order_transaction(
        self, request: CancelOrderRequest, user_id: int
    ) -> bytes:
        pass

    @abstractmethod
    def create_withdraw_transaction(
        self, request: WithdrawRequest, nonce: int, user_id: int
    ) -> bytes:
        pass

    def _create_register_message(self) -> bytes:
        message = "Welcome to ZEX."
        message = "".join(
            ("\x19Ethereum Signed Message:\n", str(len(message)), message)
        )
        return message.encode("ascii")



from typing import Literal

from pydantic import BaseModel, Field


class TradeInfo(BaseModel):
    amount: float = Field(..., description="Amount of the base token traded.")
    base_token: str = Field(..., description="Symbol of the base token.")
    id: int = Field(..., description="The ID of the trade.")
    maker_order_id: int = Field(..., description="The ID of the corresponding order.")
    name: Literal["buy", "sell"] = Field(
        ..., description="Whether it was a buy or sell trade."
    )
    price: float = Field(..., description="Trade execution price per unit.")
    quote_token: str = Field(..., description="Symbol of the quote token.")
    t: float = Field(..., description="Timestamp of the trade.")  # noqa: VNE001
    taker_order_id: int = Field(
        ..., description="The ID of the taker order of this trade."
    )



import asyncio

import pytest

from zex.sdk.client import AsyncClient
from zex.sdk.data_types import CancelOrderRequest, OrderSide, PlaceOrderRequest
from zex.sdk.websocket import ParsedWebSocketOrderMessage, ZexSocketManager


@pytest.mark.asyncio
async def test_given_valid_api_key_when_client_is_created_then_user_id_is_registered(
    zex_dev_api_key: str,
) -> None:
    # Given (Empty)  # noqa: E800
    # When: Creating a client
    client = await AsyncClient.create(api_key=zex_dev_api_key)

    # Then: The user_id should be registered and not None
    assert client.user_id is not None


@pytest.mark.asyncio
async def test_given_registered_client_when_place_empty_order_list_then_nothing_is_submitted(
    zex_dev_api_key: str,
) -> None:
    # Given: A registered client
    client = await AsyncClient.create(api_key=zex_dev_api_key)

    # When: Placing a batch with no orders
    await client.place_batch_order([])

    # Then (Empty)  # noqa: E800
    # No-op with no orders should not fail


@pytest.mark.asyncio
async def test_given_registered_client_when_cancel_empty_signed_order_list_then_nothing_is_cancelled(
    zex_dev_api_key: str,
) -> None:
    # Given: A registered client
    client = await AsyncClient.create(api_key=zex_dev_api_key)

    # When: Cancelling with no signed orders
    await client.cancel_batch_order([])

    # Then (Empty)  # noqa: E800
    # No exception should be raised and the call completes


@pytest.mark.parametrize(
    ("base_token", "quote_token", "side", "volume", "price", "zex_api_key", "testnet"),
    [
        # NOTE: The prices are too high/low so they won't get filled.
        # Dev version.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.0001,
            30000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # Regular BTC buy.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.SELL,
            0.0001,
            300000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # Regular BTC sell.
        pytest.param(
            "ETH",
            "zUSDT",
            OrderSide.BUY,
            0.0001,
            3000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # Regular ETH buy.
        pytest.param(
            "ETH",
            "zUSDT",
            OrderSide.SELL,
            0.0001,
            300000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # Regular ETH sell.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.00001,
            30000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # 5 decimal digit volume.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.000171,
            30000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # 6 Non-zero decimals.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.0001711,
            30000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # 7 Non-zero decimals.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.00017112,
            30000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # 8 Non-zero decimals.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.000171123,
            30000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # 9 Non-zero decimals.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.0001711231,
            30000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # 10 Non-zero decimals.
        pytest.param(
            "ETH",
            "zUSDT",
            OrderSide.SELL,
            0.000171123,
            300000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # Non-zero decimals ETH.
        pytest.param(
            "ETH",
            "zUSDT",
            OrderSide.SELL,
            0.000171120,
            300000.0,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # With leading zero.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.001,
            30000,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # Integer price.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.001,
            30000.01,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # Price with 2 digits.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.001,
            30000.10,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # Price Digits leading Zero.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.001,
            30000.1123,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # Price with 4 digits.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.001,
            30000.0001,
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),  # Price with small digits.
        # Main version.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.0001,
            30000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # Regular BTC buy.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.SELL,
            0.0001,
            300000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # Regular BTC sell.
        pytest.param(
            "ETH",
            "zUSDT",
            OrderSide.BUY,
            0.0001,
            3000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # Regular ETH buy.
        pytest.param(
            "ETH",
            "zUSDT",
            OrderSide.SELL,
            0.0001,
            300000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # Regular ETH sell.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.00001,
            30000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # 5 decimal digit volume.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.000171,
            30000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # 6 Non-zero decimals.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.0001711,
            30000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # 7 Non-zero decimals.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.00017112,
            30000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # 8 Non-zero decimals.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.000171123,
            30000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # 9 Non-zero decimals.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.0001711231,
            30000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # 10 Non-zero decimals.
        pytest.param(
            "ETH",
            "zUSDT",
            OrderSide.SELL,
            0.000171123,
            300000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # Non-zero decimals ETH.
        pytest.param(
            "ETH",
            "zUSDT",
            OrderSide.SELL,
            0.000171120,
            300000.0,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # With leading zero.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.001,
            30000,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # Integer price.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.001,
            30000.01,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # Price with 2 digits.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.001,
            30000.10,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # Price Digits leading Zero.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.001,
            30000.1123,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # Price with 4 digits.
        pytest.param(
            "BTC",
            "zUSDT",
            OrderSide.BUY,
            0.001,
            30000.0001,
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),  # Price with small digits.
    ],
)
@pytest.mark.asyncio
async def test_given_registered_client_when_place_and_cancel_orders_then_feedbacks_should_be_receievd_via_websocket(
    base_token: str,
    quote_token: str,
    side: OrderSide,
    volume: float,
    price: float,
    zex_api_key: str,
    testnet: bool,
) -> None:
    # Given: A registered client
    client = await AsyncClient.create(api_key=zex_api_key, testnet=testnet)
    socket_manager = ZexSocketManager(client)
    order = PlaceOrderRequest(
        base_token=base_token,
        quote_token=quote_token,
        side=side,
        volume=volume,
        price=price,
    )

    updated_order_status = []

    async def extract_new_order_status(
        order_update_message: ParsedWebSocketOrderMessage,
    ) -> None:
        updated_order_status.append(order_update_message.order_status_raw)

    # When: Placing a batch with one order and canceling that same batch
    execution_report_socket = await socket_manager.execution_report_socket(
        callback=extract_new_order_status
    )
    async with execution_report_socket:
        place_order_results = await client.place_batch_order([order])
        await asyncio.sleep(3)
        await client.cancel_batch_order(
            CancelOrderRequest(
                signed_order=place_order_result.signed_order_transaction,
                order_nonce=place_order_result.nonce,
            )
            for place_order_result in place_order_results
        )
        await asyncio.sleep(5)

    first_status = updated_order_status[0]
    second_status = updated_order_status[1]

    # Then: New order status should arrive
    assert first_status == "NEW"
    assert second_status == "CANCELED"


@pytest.mark.parametrize(
    ("zex_api_key", "testnet"),
    [
        pytest.param(
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),
        pytest.param(
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),
    ],
)
@pytest.mark.asyncio
async def test_given_a_batch_of_orders_when_place_and_cancel_then_feedbacks_should_be_received_via_websocket(
    zex_api_key: str,
    testnet: bool,
) -> None:
    # Given: A registered client
    client = await AsyncClient.create(api_key=zex_api_key, testnet=testnet)
    socket_manager = ZexSocketManager(client)
    place_order_requests = [
        PlaceOrderRequest(
            base_token="BTC",
            quote_token="zUSDT",
            side=OrderSide.BUY,
            volume=0.0001,
            price=30000,
        ),
        PlaceOrderRequest(
            base_token="BTC",
            quote_token="zUSDT",
            side=OrderSide.BUY,
            volume=0.0001,
            price=31000,
        ),
        PlaceOrderRequest(
            base_token="BTC",
            quote_token="zUSDT",
            side=OrderSide.BUY,
            volume=0.0001,
            price=32000,
        ),
        PlaceOrderRequest(
            base_token="BTC",
            quote_token="zUSDT",
            side=OrderSide.BUY,
            volume=0.0001,
            price=33000,
        ),
    ]

    updated_order_status = []

    async def extract_new_order_status(
        order_update_message: ParsedWebSocketOrderMessage,
    ) -> None:
        updated_order_status.append(order_update_message.order_status_raw)

    # When: Placing a batch with one order and canceling that same batch
    execution_report_socket = await socket_manager.execution_report_socket(
        callback=extract_new_order_status
    )
    async with execution_report_socket:
        place_order_results = await client.place_batch_order(place_order_requests)
        await asyncio.sleep(3)
        await client.cancel_batch_order(
            CancelOrderRequest(
                signed_order=place_order_result.signed_order_transaction,
                order_nonce=place_order_result.nonce,
            )
            for place_order_result in place_order_results
        )
        await asyncio.sleep(5)

    first_status = updated_order_status[0]
    second_status = updated_order_status[1]
    third_status = updated_order_status[2]
    fourth_status = updated_order_status[3]
    fifth_status = updated_order_status[4]
    sixth_status = updated_order_status[5]
    seventh_status = updated_order_status[6]
    eighth_status = updated_order_status[7]

    # Then: New order status should arrive
    assert first_status == "NEW"
    assert second_status == "NEW"
    assert third_status == "NEW"
    assert fourth_status == "NEW"
    assert fifth_status == "CANCELED"
    assert sixth_status == "CANCELED"
    assert seventh_status == "CANCELED"
    assert eighth_status == "CANCELED"


@pytest.mark.parametrize(
    ("zex_api_key", "testnet"),
    [
        pytest.param(
            pytest.lazy_fixture("zex_main_api_key"),  # type: ignore
            False,
            id="mainnet",
            marks=pytest.mark.skip(reason="Skipping the mainnet case for now."),
        ),
        pytest.param(
            pytest.lazy_fixture("zex_dev_api_key"),  # type: ignore
            True,
            id="testnet",
        ),
    ],
)
@pytest.mark.asyncio
async def test_given_a_batch_of_orders_when_placing_orders_then_order_data_should_be_retrieved_from_server(
    zex_api_key: str,
    testnet: bool,
) -> None:
    # Given: A registered client
    client = await AsyncClient.create(api_key=zex_api_key, testnet=testnet)
    socket_manager = ZexSocketManager(client)
    place_order_requests = [
        PlaceOrderRequest(
            base_token="BTC",
            quote_token="zUSDT",
            side=OrderSide.BUY,
            volume=0.0001,
            price=30000,
        ),
        PlaceOrderRequest(
            base_token="BTC",
            quote_token="zUSDT",
            side=OrderSide.BUY,
            volume=0.0001,
            price=31000,
        ),
        PlaceOrderRequest(
            base_token="BTC",
            quote_token="zUSDT",
            side=OrderSide.BUY,
            volume=0.0001,
            price=32000,
        ),
        PlaceOrderRequest(
            base_token="BTC",
            quote_token="zUSDT",
            side=OrderSide.BUY,
            volume=0.0001,
            price=33000,
        ),
    ]

    updated_order_status = []

    async def extract_new_order_status(
        order_update_message: ParsedWebSocketOrderMessage,
    ) -> None:
        updated_order_status.append(order_update_message.order_status_raw)

    # When: Placing a batch with one order and canceling that same batch
    execution_report_socket = await socket_manager.execution_report_socket(
        callback=extract_new_order_status
    )
    async with execution_report_socket:
        place_order_results = await client.place_batch_order(place_order_requests)
        await asyncio.sleep(3)
        orders = await client.get_user_orders()
        await client.cancel_batch_order(
            CancelOrderRequest(
                signed_order=place_order_result.signed_order_transaction,
                order_nonce=place_order_result.nonce,
            )
            for place_order_result in place_order_results
        )
        await asyncio.sleep(5)

    # To match with corresponding place order request.
    orders = sorted(orders, key=lambda order: order.price)

    # Then: New order status should arrive
    assert len(orders) == len(place_order_requests)
    assert orders[0].amount == place_order_requests[0].volume
    assert orders[0].base_token == place_order_requests[0].base_token
    assert orders[0].quote_token == place_order_requests[0].quote_token
    assert orders[0].name == place_order_requests[0].side.value.lower()
    assert orders[0].price == place_order_requests[0].price

    assert orders[1].amount == place_order_requests[1].volume
    assert orders[1].base_token == place_order_requests[1].base_token
    assert orders[1].quote_token == place_order_requests[1].quote_token
    assert orders[1].name == place_order_requests[1].side.value.lower()
    assert orders[1].price == place_order_requests[1].price

    assert orders[2].amount == place_order_requests[2].volume
    assert orders[2].base_token == place_order_requests[2].base_token
    assert orders[2].quote_token == place_order_requests[2].quote_token
    assert orders[2].name == place_order_requests[2].side.value.lower()
    assert orders[2].price == place_order_requests[2].price

    assert orders[3].amount == place_order_requests[3].volume
    assert orders[3].base_token == place_order_requests[3].base_token
    assert orders[3].quote_token == place_order_requests[3].quote_token
    assert orders[3].name == place_order_requests[3].side.value.lower()
    assert orders[3].price == place_order_requests[3].price
